from typing import Any


def yaml_dump(obj: Any, key: str = "") -> str:
    """
    Convert any Python object to a YAML string representation.

    Args:
        obj: Any Python object to convert to YAML

    Returns:
        str: YAML string representation of the object

    Rules:
    - Any non-first level multiline string should be rendered as block (using `|`)
    - None values are rendered correctly (not omitted)
    - Non-primitive/list/dict/set objects are ignored
    """
    import yaml

    # Process the object
    processed_obj = _sanitize_obj(obj)
    if key:
        key_parts = _parse_key(key)
        obj_to_dump = _get_obj_value(processed_obj, key_parts)
    else:
        obj_to_dump = processed_obj
    # Add custom representer for multiline strings
    yaml.add_representer(str, _multiline_string_presenter)
    # Generate YAML
    yaml_str = yaml.dump(
        obj_to_dump,
        default_flow_style=False,
        allow_unicode=True,
        sort_keys=False,
        explicit_end=False,
        width=float("inf"),
    )
    if not isinstance(obj_to_dump, (dict, list)):
        # PyYAML appends '...\n' (document-end) for top-level scalars.
        # So, we remove it.
        if yaml_str.endswith("...\n"):
            yaml_str = yaml_str[:-4]
    return yaml_str


def edit_obj(obj: Any, key: str, val: str) -> Any:
    """
    Edit a property or subproperty of an object using YAML syntax.

    Args:
        obj: The object to edit
        key: The key to edit, can be nested with '.' as separator
        val: The string value to set, will be parsed as YAML

    Returns:
        Any: The modified object

    Example:
        edit({"a": {"b": 1}}, "a.b", "2") -> {"a": {"b": 2}}
        edit({"flag": False}, "flag", "true") -> {"flag": True}
        edit({"a": 1}, "", "2") -> 2  # Replace entire object with scalar
        edit({"a": 1}, "", "b: 2") -> {"a": 1, "b": 2}  # Patch dict if obj is dict
    """
    # Parse the value using YAML rules
    parsed_value = _load_yaml(val)

    # Handle empty key - replace entire object
    if not key:
        if isinstance(obj, dict) and isinstance(parsed_value, dict):
            # Patch/merge the dict values
            return {**obj, **parsed_value}
        # Replace entire object with parsed value
        return parsed_value

    # Split the key by dots
    key_parts = _parse_key(key)
    # Set the nested value
    return _set_obj_value(obj, key_parts, parsed_value)


def _sanitize_obj(obj: Any) -> Any:
    """Process a value for YAML conversion."""
    if obj is None:
        return None
    elif isinstance(obj, (int, float, bool, str)):
        return obj
    elif isinstance(obj, (list, tuple)):
        return [_sanitize_obj(item) for item in obj if not _is_complex_obj(item)]
    elif isinstance(obj, dict):
        return {k: _sanitize_obj(v) for k, v in obj.items() if not _is_complex_obj(v)}
    elif isinstance(obj, set):
        return [
            _sanitize_obj(item) for item in sorted(obj) if not _is_complex_obj(item)
        ]
    else:
        # Ignore non-primitive/list/dict/set objects
        return None


def _is_complex_obj(obj: Any) -> bool:
    return obj is not None and not isinstance(
        obj, (int, float, bool, str, list, tuple, dict, set)
    )


def _multiline_string_presenter(dumper, data):
    """Custom representer for multiline strings."""
    if "\n" in data:
        # Clean up the string for block style
        lines = [line.rstrip() for line in data.splitlines()]
        clean_data = "\n".join(lines)
        return dumper.represent_scalar("tag:yaml.org,2002:str", clean_data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def _parse_key(key: str) -> list[str]:
    return key.split(".")


def _load_yaml(value_str: str) -> Any:
    """Parse a string value using YAML rules."""
    import yaml

    # Handle empty string explicitly
    if value_str == "":
        return ""
    try:
        # Use yaml.safe_load to parse the value
        parsed = yaml.safe_load(value_str)
        return parsed
    except yaml.YAMLError:
        # If YAML parsing fails, treat as string
        return value_str


def _set_obj_value(obj: Any, keys: list[str], value: Any) -> Any:
    """Set a value in a nested structure."""
    if not keys:
        return value
    current_key = keys[0]
    remaining_keys = keys[1:]
    if isinstance(obj, dict):
        # Handle dictionary
        if remaining_keys:
            # There are more keys to traverse
            if current_key not in obj:
                obj[current_key] = {}
            obj[current_key] = _set_obj_value(obj[current_key], remaining_keys, value)
        else:
            # This is the final key
            obj[current_key] = value
        return obj
    elif isinstance(obj, list):
        # Handle list - convert key to index
        try:
            index = int(current_key)
            if 0 <= index < len(obj):
                if remaining_keys:
                    obj[index] = _set_obj_value(obj[index], remaining_keys, value)
                else:
                    obj[index] = value
            else:
                raise IndexError(
                    f"Index {index} out of range for list of length {len(obj)}"
                )
        except ValueError:
            raise KeyError(f"Cannot use non-integer key '{current_key}' with list")
        return obj
    else:
        # Handle other types by converting to dict
        if remaining_keys:
            # Create nested structure
            new_obj = {current_key: _set_obj_value({}, remaining_keys, value)}
            return new_obj
        else:
            # Replace the entire object
            return {current_key: value}


def _get_obj_value(obj: Any, keys: list[str]) -> Any:
    """
    Get a value from a nested structure using a list of keys.
    Returns None if the key path does not exist.
    """
    current_val = obj
    for key in keys:
        if isinstance(current_val, dict):
            if key in current_val:
                current_val = current_val[key]
            else:
                return None
        elif isinstance(current_val, list):
            try:
                index = int(key)
                if 0 <= index < len(current_val):
                    current_val = current_val[index]
                else:
                    return None
            except (ValueError, TypeError):
                return None
        else:
            return None
    return current_val



import inspect
import json
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from zrb.config.llm_rate_limitter import LLMRateLimiter, llm_rate_limitter
from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.task.llm.error import extract_api_error_details
from zrb.task.llm.print_node import print_node
from zrb.task.llm.tool_wrapper import wrap_func, wrap_tool
from zrb.task.llm.typing import ListOfDict
from zrb.util.cli.style import stylize_faint

if TYPE_CHECKING:
    from pydantic_ai import Agent, Tool
    from pydantic_ai.agent import AgentRun
    from pydantic_ai.messages import UserContent
    from pydantic_ai.models import Model
    from pydantic_ai.output import OutputDataT, OutputSpec
    from pydantic_ai.settings import ModelSettings
    from pydantic_ai.toolsets import AbstractToolset

    ToolOrCallable = Tool | Callable


def create_agent_instance(
    ctx: AnyContext,
    model: "str | Model",
    output_type: "OutputSpec[OutputDataT]" = str,
    system_prompt: str = "",
    model_settings: "ModelSettings | None" = None,
    tools: "list[ToolOrCallable]" = [],
    toolsets: list["AbstractToolset[None]"] = [],
    retries: int = 3,
    yolo_mode: bool | list[str] | None = None,
) -> "Agent[None, Any]":
    """Creates a new Agent instance with configured tools and servers."""
    from pydantic_ai import Agent, RunContext, Tool
    from pydantic_ai.tools import GenerateToolJsonSchema
    from pydantic_ai.toolsets import ToolsetTool, WrapperToolset

    @dataclass
    class ConfirmationWrapperToolset(WrapperToolset):
        ctx: AnyContext
        yolo_mode: bool | list[str]

        async def call_tool(
            self, name: str, tool_args: dict, ctx: RunContext, tool: ToolsetTool[None]
        ) -> Any:
            # The `tool` object is passed in. Use it for inspection.
            # Define a temporary function that performs the actual tool call.
            async def execute_delegated_tool_call(**params):
                # Pass all arguments down the chain.
                return await self.wrapped.call_tool(name, tool_args, ctx, tool)

            # For the confirmation UI, make our temporary function look like the real one.
            try:
                execute_delegated_tool_call.__name__ = name
                execute_delegated_tool_call.__doc__ = tool.function.__doc__
                execute_delegated_tool_call.__signature__ = inspect.signature(
                    tool.function
                )
            except (AttributeError, TypeError):
                pass  # Ignore if we can't inspect the original function
            # Use the existing wrap_func to get the confirmation logic
            wrapped_executor = wrap_func(
                execute_delegated_tool_call, self.ctx, self.yolo_mode
            )
            # Call the wrapped executor. This will trigger the confirmation prompt.
            return await wrapped_executor(**tool_args)

    if yolo_mode is None:
        yolo_mode = False
    # Normalize tools
    tool_list = []
    for tool_or_callable in tools:
        if isinstance(tool_or_callable, Tool):
            tool_list.append(tool_or_callable)
            # Update tool's function
            tool = tool_or_callable
            tool_list.append(
                Tool(
                    function=wrap_func(tool.function, ctx, yolo_mode),
                    takes_ctx=tool.takes_ctx,
                    max_retries=tool.max_retries,
                    name=tool.name,
                    description=tool.description,
                    prepare=tool.prepare,
                    docstring_format=tool.docstring_format,
                    require_parameter_descriptions=tool.require_parameter_descriptions,
                    schema_generator=GenerateToolJsonSchema,
                    strict=tool.strict,
                )
            )
        else:
            # Turn function into tool
            tool_list.append(wrap_tool(tool_or_callable, ctx, yolo_mode))
    # Wrap toolsets
    wrapped_toolsets = [
        ConfirmationWrapperToolset(wrapped=toolset, ctx=ctx, yolo_mode=yolo_mode)
        for toolset in toolsets
    ]
    # Return Agent
    return Agent[None, Any](
        model=model,
        output_type=output_type,
        instructions=system_prompt,
        tools=tool_list,
        toolsets=wrapped_toolsets,
        model_settings=model_settings,
        retries=retries,
    )


def get_agent(
    ctx: AnyContext,
    model: "str | Model",
    output_type: "OutputSpec[OutputDataT]" = str,
    system_prompt: str = "",
    model_settings: "ModelSettings | None" = None,
    tools_attr: (
        "list[ToolOrCallable] | Callable[[AnyContext], list[ToolOrCallable]]"
    ) = [],
    additional_tools: "list[ToolOrCallable]" = [],
    toolsets_attr: "list[AbstractToolset[None] | str] | Callable[[AnyContext], list[AbstractToolset[None] | str]]" = [],  # noqa
    additional_toolsets: "list[AbstractToolset[None] | str]" = [],
    retries: int = 3,
    yolo_mode: bool | list[str] | None = None,
) -> "Agent":
    """Retrieves the configured Agent instance or creates one if necessary."""
    # Get tools for agent
    tools = list(tools_attr(ctx) if callable(tools_attr) else tools_attr)
    tools.extend(additional_tools)
    # Get Toolsets for agent
    toolset_or_str_list = list(
        toolsets_attr(ctx) if callable(toolsets_attr) else toolsets_attr
    )
    toolset_or_str_list.extend(additional_toolsets)
    toolsets = _render_toolset_or_str_list(ctx, toolset_or_str_list)
    # If no agent provided, create one using the configuration
    return create_agent_instance(
        ctx=ctx,
        model=model,
        output_type=output_type,
        system_prompt=system_prompt,
        tools=tools,
        toolsets=toolsets,
        model_settings=model_settings,
        retries=retries,
        yolo_mode=yolo_mode,
    )


def _render_toolset_or_str_list(
    ctx: AnyContext, toolset_or_str_list: list["AbstractToolset[None] | str"]
) -> list["AbstractToolset[None]"]:
    from pydantic_ai.mcp import load_mcp_servers

    toolsets = []
    for toolset_or_str in toolset_or_str_list:
        if isinstance(toolset_or_str, str):
            try:
                servers = load_mcp_servers(toolset_or_str)
                for server in servers:
                    toolsets.append(server)
            except Exception as e:
                ctx.log_error(f"Invalid MCP Config {toolset_or_str}: {e}")
            continue
        toolsets.append(toolset_or_str)
    return toolsets


async def run_agent_iteration(
    ctx: AnyContext,
    agent: "Agent[None, Any]",
    user_prompt: str,
    attachments: "list[UserContent] | None" = None,
    history_list: ListOfDict | None = None,
    rate_limitter: LLMRateLimiter | None = None,
    max_retry: int = 2,
    log_indent_level: int = 0,
) -> "AgentRun":
    """
    Runs a single iteration of the agent execution loop.

    Args:
        ctx: The task context.
        agent: The Pydantic AI agent instance.
        user_prompt: The user's input prompt.
        history_list: The current conversation history.

    Returns:
        The agent run result object.

    Raises:
        Exception: If any error occurs during agent execution.
    """
    if max_retry < 0:
        raise ValueError("Max retry cannot be less than 0")
    attempt = 0
    while attempt < max_retry:
        try:
            return await _run_single_agent_iteration(
                ctx=ctx,
                agent=agent,
                user_prompt=user_prompt,
                attachments=[] if attachments is None else attachments,
                history_list=[] if history_list is None else history_list,
                rate_limitter=(
                    llm_rate_limitter if rate_limitter is None else rate_limitter
                ),
                log_indent_level=log_indent_level,
            )
        except BaseException:
            attempt += 1
            if attempt == max_retry:
                raise
    raise Exception("Max retry exceeded")


async def _run_single_agent_iteration(
    ctx: AnyContext,
    agent: "Agent",
    user_prompt: str,
    attachments: "list[UserContent]",
    history_list: ListOfDict,
    rate_limitter: LLMRateLimiter,
    log_indent_level: int,
) -> "AgentRun":
    from openai import APIError
    from pydantic_ai.messages import ModelMessagesTypeAdapter

    agent_payload = _estimate_request_payload(
        agent, user_prompt, attachments, history_list
    )
    callback = _create_print_throttle_notif(ctx)
    if rate_limitter:
        await rate_limitter.throttle(agent_payload, callback)
    else:
        await llm_rate_limitter.throttle(agent_payload, callback)
    user_prompt_with_attachments = [user_prompt] + attachments
    async with agent:
        async with agent.iter(
            user_prompt=user_prompt_with_attachments,
            message_history=ModelMessagesTypeAdapter.validate_python(history_list),
        ) as agent_run:
            async for node in agent_run:
                # Each node represents a step in the agent's execution
                try:
                    await print_node(
                        _get_plain_printer(ctx), agent_run, node, log_indent_level
                    )
                except APIError as e:
                    # Extract detailed error information from the response
                    error_details = extract_api_error_details(e)
                    ctx.log_error(f"API Error: {error_details}")
                    raise
                except Exception as e:
                    ctx.log_error(f"Error processing node: {str(e)}")
                    ctx.log_error(f"Error type: {type(e).__name__}")
                    raise
            return agent_run


def _create_print_throttle_notif(ctx: AnyContext) -> Callable[[], None]:
    def _print_throttle_notif():
        ctx.print(stylize_faint("  ⌛>> Request Throttled"), plain=True)

    return _print_throttle_notif


def _estimate_request_payload(
    agent: "Agent",
    user_prompt: str,
    attachments: "list[UserContent]",
    history_list: ListOfDict,
) -> str:
    system_prompts = agent._system_prompts if hasattr(agent, "_system_prompts") else ()
    return json.dumps(
        [
            {"role": "system", "content": "\n".join(system_prompts)},
            *history_list,
            {"role": "user", "content": user_prompt},
            *[_estimate_attachment_payload(attachment) for attachment in attachments],
        ]
    )


def _estimate_attachment_payload(attachment: "UserContent") -> Any:
    if hasattr(attachment, "url"):
        return {"role": "user", "content": attachment.url}
    if hasattr(attachment, "data"):
        return {"role": "user", "content": "x" * len(attachment.data)}
    return ""


def _get_plain_printer(ctx: AnyContext):
    def printer(*args, **kwargs):
        if "plain" not in kwargs:
            kwargs["plain"] = True
        return ctx.print(*args, **kwargs)

    return printer



from typing import Any, Callable

from zrb.context.any_context import AnyContext

fstring = str
AnyAttr = Any | fstring | Callable[[AnyContext], Any]
StrAttr = str | fstring | Callable[[AnyContext], str | None]
BoolAttr = bool | fstring | Callable[[AnyContext], bool | None]
IntAttr = int | fstring | Callable[[AnyContext], int | None]
FloatAttr = float | fstring | Callable[[AnyContext], float | None]
StrDictAttr = dict[str, StrAttr] | Callable[[AnyContext], dict[str, Any]]
StrListAttr = list[StrAttr] | Callable[[AnyContext], list[str]]



import subprocess
import sys

if sys.version_info >= (3, 12):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class ShellCommandResult(TypedDict):
    """
    Result of shell command execution

    Attributes:
        return_code: The return code, 0 indicating no error
        stdout: Standard output
        stderr: Standard error
    """

    return_code: int
    stdout: str
    stderr: str


def run_shell_command(command: str) -> ShellCommandResult:
    """
    Executes a non-interactive shell command on the user's machine.

    CRITICAL: This tool runs with user-level permissions. Explain commands that modify
        the system (e.g., `git`, `pip`) and ask for confirmation.
    IMPORTANT: Long-running processes should be run in the background (e.g., `command &`).

    Example:
    run_shell_command(command='ls -l')

    Args:
        command (str): The exact shell command to be executed.

    Returns:
        dict: return_code, stdout, and stderr.
    """
    result = subprocess.run(
        command,
        shell=True,
        capture_output=True,
        text=True,
    )
    return {
        "return_code": result.returncode,
        "stdout": result.stdout,
        "stderr": result.stderr,
    }



from abc import ABC, abstractmethod
from typing import Any

from zrb.session.any_session import AnySession


class AnyCallback(ABC):
    @abstractmethod
    async def async_run(self, parent_session: AnySession, session: AnySession) -> Any:
        pass



import datetime
from typing import Any

import ulid
from my_app_name.common.base_db_repository import BaseDBRepository
from my_app_name.common.error import InvalidValueError, NotFoundError, UnauthorizedError
from my_app_name.module.auth.service.user.repository.user_repository import (
    UserRepository,
)
from my_app_name.schema.permission import Permission
from my_app_name.schema.role import Role, RolePermission
from my_app_name.schema.user import (
    User,
    UserCreateWithAudit,
    UserResponse,
    UserRole,
    UserSession,
    UserSessionResponse,
    UserTokenData,
    UserUpdateWithAudit,
)
from passlib.context import CryptContext
from sqlalchemy.sql import Select
from sqlmodel import delete, insert, select, update

# Password hashing context
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def hash_password(password: str) -> str:
    return pwd_context.hash(password)


def verify_password(plain_password: str, hashed_password: str) -> bool:
    """Verifies if a password matches the stored hash."""
    return pwd_context.verify(plain_password, hashed_password)


class UserDBRepository(
    BaseDBRepository[User, UserResponse, UserCreateWithAudit, UserUpdateWithAudit],
    UserRepository,
):
    db_model = User
    response_model = UserResponse
    create_model = UserCreateWithAudit
    update_model = UserUpdateWithAudit
    entity_name = "user"
    column_preprocessors = {"password": hash_password}

    def _select(self) -> Select:
        return (
            select(User, Role, Permission)
            .join(UserRole, UserRole.user_id == User.id, isouter=True)
            .join(Role, Role.id == UserRole.role_id, isouter=True)
            .join(RolePermission, RolePermission.role_id == Role.id, isouter=True)
            .join(
                Permission, Permission.id == RolePermission.permission_id, isouter=True
            )
        )

    def _rows_to_responses(self, rows: list[tuple[Any, ...]]) -> list[UserResponse]:
        user_map: dict[str, dict[str, Any]] = {}
        user_role_map: dict[str, list[str]] = {}
        user_permission_map: dict[str, list[str]] = {}
        for user, role, permission in rows:
            if user.id not in user_map:
                user_map[user.id] = {"user": user, "roles": [], "permissions": []}
                user_role_map[user.id] = []
                user_permission_map[user.id] = []
            if role is not None and role.id not in user_role_map[user.id]:
                user_role_map[user.id].append(role.id)
                user_map[user.id]["roles"].append(role)
            if (
                permission is not None
                and permission.id not in user_permission_map[user.id]
            ):
                user_permission_map[user.id].append(permission.id)
                user_map[user.id]["permissions"].append(permission)
        return [
            UserResponse(
                **data["user"].model_dump(),
                role_names=[role.name for role in data["roles"]],
                permission_names=[
                    permission.name for permission in data["permissions"]
                ],
            )
            for data in user_map.values()
        ]

    async def validate_role_names(self, role_names: list[str]):
        async with self._session_scope() as session:
            result = await self._execute_statement(
                session, select(Role.name).where(Role.name.in_(role_names))
            )
            existing_roles = {row[0] for row in result.all()}
            # Identify any missing role names
            missing_roles = set(role_names) - existing_roles
            if missing_roles:
                raise InvalidValueError(
                    f"Role(s) not found: {', '.join(missing_roles)}"
                )

    async def add_roles(self, data: dict[str, list[str]], created_by: str):
        now = datetime.datetime.now(datetime.timezone.utc)
        # get mapping from role names to role ids
        all_role_names = {name for role_names in data.values() for name in role_names}
        async with self._session_scope() as session:
            result = await self._execute_statement(
                session, select(Role.id, Role.name).where(Role.name.in_(all_role_names))
            )
            role_mapping = {row.name: row.id for row in result}
        # Assemble data dict
        data_dict_list: list[dict[str, Any]] = []
        for user_id, role_names in data.items():
            for role_name in role_names:
                data_dict_list.append(
                    self._model_to_data_dict(
                        UserRole(
                            id=ulid.new().str,
                            user_id=user_id,
                            role_id=role_mapping.get(role_name),
                            created_at=now,
                            created_by=created_by,
                        )
                    )
                )
        if len(data_dict_list) == 0:
            return
        async with self._session_scope() as session:
            await self._execute_statement(
                session, insert(UserRole).values(data_dict_list)
            )

    async def remove_all_roles(self, user_ids: list[str] = []):
        async with self._session_scope() as session:
            await self._execute_statement(
                session,
                delete(UserRole).where(UserRole.user_id.in_(user_ids)),
            )

    async def get_by_credentials(self, username: str, password: str) -> UserResponse:
        async with self._session_scope() as session:
            result = await self._execute_statement(
                session, select(User).where(User.username == username, User.active)
            )
            user = result.scalar_one_or_none()
            if user is None or not verify_password(password, user.password):
                raise UnauthorizedError("Invalid username or password")
            return await self.get_by_id(user.id)

    async def delete_expired_user_sessions(self, user_id: str):
        now = datetime.datetime.now(datetime.timezone.utc)
        async with self._session_scope() as session:
            await self._execute_statement(
                session,
                delete(UserSession).where(
                    UserSession.user_id == user_id,
                    UserSession.refresh_token_expired_at < now,
                ),
            )

    async def get_active_user_sessions(self, user_id: str) -> list[UserSessionResponse]:
        now = datetime.datetime.now(datetime.timezone.utc)
        async with self._session_scope() as session:
            result = await self._execute_statement(
                session,
                select(UserSession).where(
                    UserSession.user_id == user_id,
                    UserSession.refresh_token_expired_at > now,
                ),
            )
            return [self._user_session_to_response(row[0]) for row in result.all()]

    async def get_user_session_by_access_token(
        self, access_token: str
    ) -> UserSessionResponse:
        now = datetime.datetime.now(datetime.timezone.utc)
        async with self._session_scope() as session:
            result = await self._execute_statement(
                session,
                select(UserSession).where(
                    UserSession.access_token == access_token,
                    UserSession.access_token_expired_at > now,
                ),
            )
            user_session = result.scalar_one_or_none()
            if user_session is None:
                raise NotFoundError("User session not found")
            return self._user_session_to_response(user_session)

    async def get_user_session_by_refresh_token(
        self, refresh_token: str
    ) -> UserSessionResponse:
        now = datetime.datetime.now(datetime.timezone.utc)
        async with self._session_scope() as session:
            result = await self._execute_statement(
                session,
                select(UserSession).where(
                    UserSession.refresh_token == refresh_token,
                    UserSession.refresh_token_expired_at > now,
                ),
            )
            user_session = result.scalar_one_or_none()
            if user_session is None:
                raise NotFoundError("User session not found")
            return self._user_session_to_response(user_session)

    async def create_user_session(
        self, user_id: str, token_data: UserTokenData
    ) -> UserSessionResponse:
        data_dict = self._model_to_data_dict(
            token_data, user_id=user_id, id=ulid.new().str
        )
        async with self._session_scope() as session:
            await self._execute_statement(
                session, insert(UserSession).values(**data_dict)
            )
            result = await self._execute_statement(
                session, select(UserSession).where(UserSession.id == data_dict["id"])
            )
            user_session = result.scalar_one_or_none()
            if user_session is None:
                raise NotFoundError("User session not found after created")
            return self._user_session_to_response(user_session)

    async def update_user_session(
        self, user_id: str, session_id: str, token_data: UserTokenData
    ) -> UserSessionResponse:
        data_dict = self._model_to_data_dict(token_data, user_id=user_id)
        async with self._session_scope() as session:
            await self._execute_statement(
                session,
                (
                    update(UserSession)
                    .where(UserSession.id == session_id)
                    .values(**data_dict)
                ),
            )
            result = await self._execute_statement(
                session, select(UserSession).where(UserSession.id == session_id)
            )
            user_session = result.scalar_one_or_none()
            if user_session is None:
                raise NotFoundError("User session not found after created")
            return self._user_session_to_response(user_session)

    async def delete_user_sessions(self, session_ids: list[str]):
        async with self._session_scope() as session:
            await self._execute_statement(
                session, delete(UserSession).where(UserSession.id.in_(session_ids))
            )

    def _user_session_to_response(
        self, user_session: UserSession
    ) -> UserSessionResponse:
        return UserSessionResponse(
            id=user_session.id,
            user_id=user_session.user_id,
            access_token=user_session.access_token,
            access_token_expired_at=user_session.access_token_expired_at,
            refresh_token=user_session.refresh_token,
            refresh_token_expired_at=user_session.refresh_token_expired_at,
            token_type="bearer",
        )



import os

from zrb.context.any_shared_context import AnySharedContext
from zrb.input.str_input import StrInput
from zrb.util.file import read_file
from zrb.util.string.conversion import to_pascal_case


class PreviousSessionInput(StrInput):

    def to_html(self, ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = self.get_default_str(ctx)
        script = read_file(
            file_path=os.path.join(os.path.dirname(__file__), "previous-session.js"),
            replace_map={
                "CURRENT_INPUT_NAME": name,
                "CurrentPascalInputName": to_pascal_case(name),
            },
        )
        return "\n".join(
            [
                f'<input name="{name}" placeholder="{description}" value="{default}" />',
                f"<script>{script}</script>",
            ]
        )



from my_app_name.common.util.parser import (
    create_default_filter_param_parser,
    create_default_sort_param_parser,
)

parse_filter_param = create_default_filter_param_parser()
parse_sort_param = create_default_sort_param_parser()



import re

from zrb.util.markdown import promote_markdown_headers


def markdown_to_dict(markdown: str) -> dict[str, str]:
    sections: dict[str, str] = {}
    current_title = ""
    current_content: list[str] = []
    fence_stack: list[str] = []
    fence_pattern = re.compile(r"^([`~]{3,})(.*)$")
    h1_pattern = re.compile(r"^# (.+)$")
    for line in markdown.splitlines():
        # Detect code fence open/close
        fence_match = fence_pattern.match(line.strip())
        if fence_match:
            fence = fence_match.group(1)
            if fence_stack and fence_stack[-1] == fence:
                fence_stack.pop()  # close current fence
            else:
                fence_stack.append(fence)  # open new fence
        # Only parse H1 when not inside a code fence
        if not fence_stack:
            h1_match = h1_pattern.match(line)
            if h1_match:
                # Save previous section
                if current_title:
                    sections[current_title] = "\n".join(current_content).strip()
                # Start new section
                current_title = h1_match.group(1).strip()
                current_content = []
                continue
        current_content.append(line)
    # Save final section
    if current_title:
        sections[current_title] = "\n".join(current_content).strip()
    return {
        header: promote_markdown_headers(content)
        for header, content in sections.items()
    }



import datetime
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from zrb.session_state_log.session_state_log import (
        SessionStateLog,
        SessionStateLogList,
    )


class AnySessionStateLogger(ABC):
    @abstractmethod
    def write(self, session_log: "SessionStateLog"):
        pass

    @abstractmethod
    def read(self, session_name: str) -> "SessionStateLog":
        pass

    @abstractmethod
    def list(
        self,
        task_path: list[str],
        min_start_time: datetime.datetime,
        max_start_time: datetime.datetime,
        page: int = 0,
        limit: int = 10,
    ) -> "SessionStateLogList":
        pass



from my_app_name.schema.permission import Permission
from my_app_name.schema.role import Role, RolePermission
from my_app_name.schema.user import User, UserRole, UserSession
from sqlalchemy import MetaData

metadata = MetaData()

Permission.metadata = metadata
Permission.__table__.tometadata(metadata)

Role.metadata = metadata
Role.__table__.tometadata(metadata)
RolePermission.metadata = metadata
RolePermission.__table__.tometadata(metadata)

User.metadata = metadata
User.__table__.tometadata(metadata)
UserRole.metadata = metadata
UserRole.__table__.tometadata(metadata)

UserSession.metadata = metadata
UserSession.__table__.tometadata(metadata)



import os

from zrb.config.config import CFG
from zrb.config.llm_context.config_parser import markdown_to_dict
from zrb.config.llm_context.workflow import LLMWorkflow
from zrb.util.markdown import demote_markdown_headers


class LLMContextConfig:
    """High-level API for interacting with cascaded configurations."""

    def write_note(
        self,
        content: str,
        context_path: str | None = None,
        cwd: str | None = None,
    ):
        """Writes content to a note block in the user's home configuration file."""
        if cwd is None:
            cwd = os.getcwd()
        if context_path is None:
            context_path = cwd
        config_file = self._get_home_config_file()
        sections = {}
        if os.path.exists(config_file):
            sections = self._parse_config(config_file)
        abs_context_path = os.path.abspath(os.path.join(cwd, context_path))
        found_key = None
        for key in sections.keys():
            if not key.startswith("Note:"):
                continue
            context_path_str = key[len("Note:") :].strip()
            abs_key_path = self._normalize_context_path(
                context_path_str,
                os.path.dirname(config_file),
            )
            if abs_key_path == abs_context_path:
                found_key = key
                break
        if found_key:
            sections[found_key] = content
        else:
            config_dir = os.path.dirname(config_file)
            formatted_path = self._format_context_path_for_writing(
                abs_context_path,
                config_dir,
            )
            new_key = f"Note: {formatted_path}"
            sections[new_key] = content
        # Serialize back to markdown
        new_file_content = ""
        for key, value in sections.items():
            new_file_content += f"# {key}\n{demote_markdown_headers(value)}\n\n"
        with open(config_file, "w") as f:
            f.write(new_file_content)

    def get_notes(self, cwd: str | None = None) -> dict[str, str]:
        """Gathers all notes for a given path."""
        if cwd is None:
            cwd = os.getcwd()
        config_file = self._get_home_config_file()
        if not os.path.exists(config_file):
            return {}
        config_dir = os.path.dirname(config_file)
        sections = self._parse_config(config_file)
        notes: dict[str, str] = {}
        for key, value in sections.items():
            if key.lower().startswith("note:"):
                context_path_str = key[len("note:") :].strip()
                abs_context_path = self._normalize_context_path(
                    context_path_str,
                    config_dir,
                )
                # A context is relevant if its path is an ancestor of cwd
                if os.path.commonpath([cwd, abs_context_path]) == abs_context_path:
                    notes[abs_context_path] = value
        return notes

    def get_workflows(self, cwd: str | None = None) -> dict[str, LLMWorkflow]:
        """Gathers all relevant workflows for a given path."""
        if cwd is None:
            cwd = os.getcwd()
        all_sections = self._get_all_sections(cwd)
        workflows: dict[str, LLMWorkflow] = {}
        # Iterate from closest to farthest
        for config_dir, sections in all_sections:
            for key, value in sections.items():
                if key.lower().startswith("workflow:"):
                    workflow_name = key[len("workflow:") :].strip().lower()
                    # First one found wins
                    if workflow_name not in workflows:
                        workflows[workflow_name] = LLMWorkflow(
                            name=workflow_name,
                            content=value,
                            path=config_dir,
                        )
        return workflows

    def _format_context_path_for_writing(
        self,
        path_to_write: str,
        relative_to_dir: str,
    ) -> str:
        """Formats a path for writing into a context file key."""
        home_dir = os.path.expanduser("~")
        abs_path_to_write = os.path.abspath(
            os.path.join(relative_to_dir, path_to_write)
        )
        abs_relative_to_dir = os.path.abspath(relative_to_dir)
        # Rule 1: Inside relative_to_dir
        if abs_path_to_write.startswith(abs_relative_to_dir):
            if abs_path_to_write == abs_relative_to_dir:
                return "."
            return os.path.relpath(abs_path_to_write, abs_relative_to_dir)
        # Rule 2: Inside Home
        if abs_path_to_write.startswith(home_dir):
            if abs_path_to_write == home_dir:
                return "~"
            return os.path.join("~", os.path.relpath(abs_path_to_write, home_dir))
        # Rule 3: Absolute
        return abs_path_to_write

    def _find_config_files(self, cwd: str) -> list[str]:
        configs = []
        current_dir = cwd
        home_dir = os.path.expanduser("~")
        while True:
            config_path = os.path.join(current_dir, CFG.LLM_CONTEXT_FILE)
            if os.path.exists(config_path):
                configs.append(config_path)
            if current_dir == home_dir or current_dir == "/":
                break
            current_dir = os.path.dirname(current_dir)
        return configs

    def _get_home_config_file(self) -> str:
        home_dir = os.path.expanduser("~")
        return os.path.join(home_dir, CFG.LLM_CONTEXT_FILE)

    def _parse_config(self, file_path: str) -> dict[str, str]:
        with open(file_path, "r") as f:
            content = f.read()
        return markdown_to_dict(content)

    def _get_all_sections(self, cwd: str) -> list[tuple[str, dict[str, str]]]:
        config_files = self._find_config_files(cwd)
        all_sections = []
        for config_file in config_files:
            config_dir = os.path.dirname(config_file)
            sections = self._parse_config(config_file)
            all_sections.append((config_dir, sections))
        return all_sections

    def _normalize_context_path(
        self,
        path_str: str,
        relative_to_dir: str,
    ) -> str:
        """Normalizes a context path string to an absolute path."""
        expanded_path = os.path.expanduser(path_str)
        if os.path.isabs(expanded_path):
            return os.path.abspath(expanded_path)
        return os.path.abspath(os.path.join(relative_to_dir, expanded_path))


llm_context_config = LLMContextConfig()



import re


def get_os_from_user_agent(user_agent: str) -> str:
    os_patterns = [
        (r"Windows NT (\d+\.\d+)", "Windows"),
        (r"Mac OS X (\d+[_\.]\d+)", "MacOS"),
        (r"Android (\d+[\.\d]*)", "Android"),
        (r"iPhone OS (\d+[_\.]\d+)", "iOS"),
        (r"iPad.*OS (\d+[_\.]\d+)", "iPadOS"),
        (r"Linux", "Linux"),
    ]
    os = "Unknown OS"
    # Match OS
    for pattern, name in os_patterns:
        match = re.search(pattern, user_agent)
        if match:
            os = (
                f"{name} {match.group(1).replace('_', '.')}" if match.groups() else name
            )
            break
    return os


def get_browser_from_user_agent(user_agent: str) -> str:
    browser_patterns = [
        (r"Chrome/([\d\.]+)", "Chrome"),
        (r"Firefox/([\d\.]+)", "Firefox"),
        (r"Edg/([\d\.]+)", "Edge"),
        (r"Safari/([\d\.]+)", "Safari"),
    ]
    browser = "Unknown Browser"
    # Match Browser
    for pattern, name in browser_patterns:
        match = re.search(pattern, user_agent)
        if match:
            browser = f"{name} {match.group(1)}"
            break
    return browser


def get_device_from_user_agent(user_agent: str):
    device_patterns = [
        (r"iPhone", "iPhone"),
        (r"iPad", "iPad"),
        (r"Android.*Mobile", "Android Phone"),
        (r"Android", "Android Tablet"),
        (r"Macintosh", "Mac"),
        (r"Windows NT", "Windows PC"),
        (r"Linux", "Linux Device"),
    ]
    device = "Unknown Device"
    # Match Device
    for pattern, name in device_patterns:
        if re.search(pattern, user_agent):
            device = name
            break
    return device



import os

from zrb.config.config import CFG
from zrb.config.llm_context.config import llm_context_config
from zrb.config.llm_context.workflow import LLMWorkflow


def load_workflow(workflow_name: str | list[str]) -> str:
    """
    Loads and formats one or more workflow documents for LLM consumption.

    Retrieves workflows by name, formats with descriptive headers for LLM context injection.

    Args:
        workflow_name: Name or list of names of the workflow(s) to load

    Returns:
        Formatted workflow content as a string with headers

    Raises:
        ValueError: If any specified workflow name is not found
    """
    names = [workflow_name] if isinstance(workflow_name, str) else workflow_name
    available_workflows = get_available_workflows()
    contents = []
    for name in names:
        workflow = available_workflows.get(name.strip().lower())
        if workflow is None:
            raise ValueError(f"Workflow not found: {name}")
        contents.append(
            "\n".join(
                [
                    f"# {workflow.name}",
                    f"> Workflow Location: `{workflow.path}`",
                    workflow.content,
                ]
            )
        )
    return "\n".join(contents)


def get_available_workflows() -> dict[str, LLMWorkflow]:
    available_workflows = {
        workflow_name.strip().lower(): workflow
        for workflow_name, workflow in llm_context_config.get_workflows().items()
    }
    # Define builtin workflow locations in order of precedence
    builtin_workflow_locations = [
        os.path.expanduser(additional_builtin_workflow_path)
        for additional_builtin_workflow_path in CFG.LLM_BUILTIN_WORKFLOW_PATHS
        if os.path.isdir(os.path.expanduser(additional_builtin_workflow_path))
    ]
    builtin_workflow_locations.append(
        os.path.join(os.path.dirname(__file__), "default_workflow")
    )
    # Load workflows from all locations
    for workflow_location in builtin_workflow_locations:
        if not os.path.isdir(workflow_location):
            continue
        for workflow_name in os.listdir(workflow_location):
            workflow_dir = os.path.join(workflow_location, workflow_name)
            workflow_file = os.path.join(workflow_dir, "workflow.md")
            if not os.path.isfile(workflow_file):
                workflow_file = os.path.join(workflow_dir, "SKILL.md")
                if not os.path.isfile(path=workflow_file):
                    continue
            # Only add if not already defined (earlier locations have precedence)
            if workflow_name not in available_workflows:
                with open(workflow_file, "r") as f:
                    workflow_content = f.read()
                available_workflows[workflow_name] = LLMWorkflow(
                    name=workflow_name,
                    path=workflow_dir,
                    content=workflow_content,
                )
    return available_workflows



import asyncio
from collections.abc import Callable
from typing import Any

from zrb.attr.type import fstring
from zrb.callback.any_callback import AnyCallback
from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.context.shared_context import SharedContext
from zrb.dot_dict.dot_dict import DotDict
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.session.any_session import AnySession
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.cli.style import CYAN
from zrb.xcom.xcom import Xcom


class BaseTrigger(BaseTask):
    """
    A base class for tasks that act as triggers or schedulers.

    It extends BaseTask and adds functionality for handling callbacks
    and managing a queue for data exchange.
    """

    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        action: fstring | Callable[[AnyContext], Any] | None = None,
        execute_condition: bool | str | Callable[[AnyContext], bool] = True,
        queue_name: fstring | None = None,
        callback: list[AnyCallback] | AnyCallback = [],
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
    ):
        """
        Initializes a new instance of the BaseTrigger class.

        Args:
            name: The name of the trigger task.
            color: The color to use for the task in the CLI output.
            icon: The icon to display for the task.
            description: A brief description of the task.
            cli_only: If True, the task is only available in the CLI.
            input: The input definition for the task.
            env: The environment variable definition for the task.
            action: The action to be performed by the task.
            execute_condition: A condition that must be met for the task to execute.
            queue_name: The name of the XCom queue used for data
                exchange with callbacks. Whenever any data is added
                to xcom[queue_name], the callback will be triggered.
            callback: A single or list of callbacks to be executed after the trigger action.
            retries: The number of times to retry the task on failure.
            retry_period: The time to wait between retries.
            readiness_check: A single or list of tasks to check for readiness before execution.
            readiness_check_delay: The initial delay before starting
                readiness checks.
            readiness_check_period: The time to wait between readiness checks.
            readiness_failure_threshold: The number of consecutive readiness
                check failures before the task fails.
            readiness_timeout: The maximum time to wait for readiness checks to pass.
            monitor_readiness: If True, monitor readiness checks during execution.
            upstream: A single or list of tasks that must complete before this task starts.
            fallback: A single or list of tasks to run if this task fails.
            successor: A single or list of tasks to run after this task completes successfully.
        """
        super().__init__(
            name=name,
            color=color if color is not None else CYAN,
            icon=icon if icon is not None else "✨",
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            action=action,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
            successor=successor,
        )
        self._callbacks = callback
        self._queue_name = queue_name

    @property
    def queue_name(self) -> str:
        if self._queue_name is None:
            return f"{self.name}"
        return self._queue_name

    @property
    def readiness_checks(self) -> list[AnyTask]:
        readiness_checks = super().readiness_checks
        if len(readiness_checks) > 0:
            return readiness_checks
        return [BaseTask(name=f"{self.name}-check", action=lambda _: True)]

    @property
    def callbacks(self) -> list[AnyCallback]:
        if isinstance(self._callbacks, AnyCallback):
            return [self._callbacks]
        return self._callbacks

    async def exec_root_tasks(self, session: AnySession):
        exchange_xcom = self._get_exchange_xcom(session)
        exchange_xcom.add_push_callback(lambda: self._exchange_push_callback(session))
        return await super().exec_root_tasks(session)

    def _exchange_push_callback(self, session: AnySession):
        coro = asyncio.create_task(self._fanout_and_trigger_callback(session))
        session.defer_coro(coro)

    async def _fanout_and_trigger_callback(self, session: AnySession):
        data = self.pop_exchange_xcom(session)
        coros = []
        for callback in self.callbacks:
            xcom_dict = DotDict({self.queue_name: Xcom([data])})
            callback_session = Session(
                shared_ctx=SharedContext(
                    input=dict(session.shared_ctx.input),
                    xcom=xcom_dict,
                ),
                parent=session,
                root_group=session.root_group,
            )
            coros.append(
                asyncio.create_task(
                    callback.async_run(parent_session=session, session=callback_session)
                )
            )
        await asyncio.gather(*coros)

    def _get_exchange_xcom(self, session: AnySession) -> Xcom:
        shared_ctx = session.shared_ctx
        if self.queue_name not in shared_ctx.xcom:
            shared_ctx.xcom[self.queue_name] = Xcom()
        return shared_ctx.xcom[self.queue_name]

    def push_exchange_xcom(self, session: AnySession, data: Any):
        exchange_xcom = self._get_exchange_xcom(session)
        exchange_xcom.push(data)

    def pop_exchange_xcom(self, session: AnySession) -> Any:
        exchange_xcom = self._get_exchange_xcom(session)
        return exchange_xcom.pop()



from my_app_name.common.db_engine_factory import db_engine
from my_app_name.config import APP_REPOSITORY_TYPE
from my_app_name.module.my_module.service.my_entity.repository.my_entity_db_repository import (
    MyEntityDBRepository,
)
from my_app_name.module.my_module.service.my_entity.repository.my_entity_repository import (
    MyEntityRepository,
)

if APP_REPOSITORY_TYPE == "db":
    my_entity_repository: MyEntityRepository = MyEntityDBRepository(db_engine)
else:
    my_entity_repository: MyEntityRepository = None
